// TextConsole.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>

#define TYPE_CODE 0
#define TYPE_ARGLIST 1


struct CCNode {
	int level;		// depth, i.e., CC byte # (sorta)
	int value;		// byte value (-1 for arguments)

	int type;		// code byte or argument
	int num_arg;	// number of argument bytes
	int args[32];	// argument list
	bool arg_multiplier;	// if true, number of arguments remaining is multiplied by the
							// value of the first

	bool isTerminator;

	int num_nodes;
	CCNode *nodes;
};

// String class
struct StrInfo {
	int address;

	char* str;
	int len;
};



// global functions

void CreateCCTable(CCNode *table);
void CreateCompressionTable(FILE* rom, char* tbl[]);

CCNode* AddNode(CCNode *node, int level, int value, int type, int num_arg, bool multi, bool terminate);

StrInfo ReadString(FILE* rom, unsigned long address);
char* ParseString(char str[], int len);

int SearchNode(CCNode *node, int val);
void AddCode(char* str, int val, int pos);
bool isCC(char ch);



unsigned int HexToSNES(unsigned int);
unsigned int SNESToHex(unsigned int);

void LoadRawText(FILE* rom, StrInfo strlist[]);
void LoadTPTText();

void LoadTPT();



// global properties

char* comprTable[768];

CCNode ccTable;	// initial node

// pointer table text
StrInfo tpttext[10000];
int num_tpt;

// raw text
StrInfo rawtext[10000];
int num_raw;

//

FILE* rom;




int main(int argc, char* argv[])
{
	char filename[256];
	
	CreateCCTable(&ccTable);

	if(argc <= 1)
	{
		printf("Filename? ");
		cin >> filename;
		rom = fopen(filename, "rb");
	}
	else
	{
		rom = fopen(argv[1], "rb");
	}

	if(!rom)
	{
		printf("ERROR: failed to open file.\n");
		return -1;
	}

	CreateCompressionTable(rom, comprTable);

	LoadTPTText();
	LoadRawText(rom, rawtext);

	printf("...That sure was fast, warn't it? :D\n");


	int address = 0;

	while(address != -1)
	{
		cout << "\nEnter text index (-1 to quit): ";
		cin >> address;

		if(address == -1) break;
		if(address >= num_raw)
		{
			printf("ERROR: Index too high.\n\n");
			continue;
		}

		StrInfo s = rawtext[address];

		printf("\nText at: 0x%X ($%X)\n================\n\"%s\"\n\n[String is %d bytes long (uncompressed)]\n", s.address, HexToSNES(s.address), ParseString(s.str, s.len), s.len);

	}


	return 0;
}


void LoadRawText(FILE* rom, StrInfo strlist[])
{
	// Loads the raw text area(s) into the specified list of strings.
	// 

	int address = 0x51D12;
	int istr = 0;

	do
	{
		if(address < 0x0A012E || address >= 0x2F5020)
		{
			if(address < 0x08BE2D || address >= 0x8DC31)
			{
				StrInfo s = ReadString(rom, address);

				rawtext[istr] = s;

				address+=s.len;
				istr++;
			}
			else
				address = 0x8DC31;
		}
		else
			address = 0x2F5020;
	}while(address < 0x2FA57A);

	num_raw = istr;

	printf("Done loading raw text!\n%d strings loaded.\n\n", istr);
}



void LoadTPTText()
{
	int offset = 0;

	int istr = 0;

	int address;

	for(int i = 0; i < 1583; i++)
	{

		fseek(rom, 0xF8B9F+offset, SEEK_SET);

		address = fgetc(rom);
		address+= fgetc(rom) << 8;
		address+= fgetc(rom) << 16;

		if(address != 0)
		{
			StrInfo s = ReadString(rom, SNESToHex(address));
			tpttext[istr++] = s;
		}

		offset+=0x11;
	}
	num_tpt = istr;
	printf("Done loading TPT entries!\n%d strings loaded.\n\n", istr);
}



StrInfo ReadString(FILE* rom, unsigned long address)
{
	// Reads a string from ROM at address.
	// Returns a struct containing a pointer to the string and
	// the string's length. This function does not do any formatting
	// or decompression. If the string is greater than 8192 characters
	// long, it will be truncated at that point.

	char buffer[8192];

	int pos = 0;
	int len = 0;

	CCNode* activeNode = &ccTable;

	bool opcode = false;
	int arglevel = 0;

	unsigned char ch;

	fseek(rom, address, SEEK_SET);

	for(int i = 0; i < 8192; i++)
	{
		ch = fgetc(rom);
		len++;

		if(arglevel > 0)
		{
			buffer[pos++] = ch;

			arglevel--;
			continue;
		}

		if(!opcode)
		{
			if(isCC(ch))
			{
				opcode = true;
			}
		}

		if(opcode)
		{
			int n = SearchNode(activeNode, ch);
			if(n == -1)
			{
				opcode = false;
				activeNode = &ccTable;
				buffer[pos++] = ch;
				continue;				
			}
			else
			{
				activeNode = &activeNode->nodes[n];
			}
			
			if(activeNode->type == TYPE_ARGLIST)
			{
				arglevel = activeNode->num_arg-1;
				buffer[pos++] = ch;
			}
			else if(activeNode->type == TYPE_CODE)
			{
				buffer[pos++] = ch;
			}

			if(activeNode->level == 0 && activeNode->value == 0x02)
				break;

			if(activeNode->isTerminator)
			{
				opcode = false;
				activeNode = &ccTable;
			}
		}
		else
		{
			buffer[pos++] = ch-0x30;
		}
	}

	buffer[pos++] = 0;

	StrInfo s;
	s.address = address;
	s.str = new char[pos];
	s.len = len;
	
	for(i = 0; i < pos; i++)
		s.str[i] = buffer[i];

	return s;
}




char* ParseString(char str[], int len)
{
	char buffer[8192];

	int pos = 0;	// position in returned string
	int ipos = 0;

	CCNode* activeNode = &ccTable;

	bool opcode = false;
	int arglevel = 0;

	int curcode;		// initial byte of current control code

	unsigned char ch;

	for(int i = 0; i < len; i++)
	{
		ch = str[ipos++];
		//len++;

		// if there are arguments left to count, decrement arg level and continue
		if(arglevel > 0)
		{
			AddCode(buffer, ch, pos);
			pos+=4;

			arglevel--;
			continue;
		}

		// if opcode was not previously set
		if(!opcode)
		{
			if(isCC(ch))
			{
				curcode = ch;
				opcode = true;
			}
		}


		// do more stuff

		// if opcode is true, ...

		// note that none of this will happen when there are arguments in the queue,
		// so the active node will still be the original argument node when the arguments
		// finally close, so everything will proceed in an orderly fashion.
		// (an argument node will just run through a few extra bytes and add them to the string,
		// whereas other nodes involve adding only one character to the string)

		if(opcode)
		{

			// search for a subnode matching the current byte
			// (as of initial iteration, activeNode is the root node)
			int n = SearchNode(activeNode, ch);
			if(n == -1)
			{
				// if no match was found, this is an invalid code
				opcode = false;
				activeNode = &ccTable;
			}
			else
			{
				// set the active node to a subnode of itself
				activeNode = &activeNode->nodes[n];
			}
			

			// okay, now we have a byte from the ROM, and have associated it
			// with a node. depending on the node type, do stuff

			if(activeNode->type == TYPE_ARGLIST)
			{
				// set arg level to num_arg - 1 because we've already read
				// one of the argument bytes
				arglevel = activeNode->num_arg-1;

				if(curcode == 0x15 || curcode == 0x16 || curcode == 0x17)
				{
					int c = ((curcode-0x15)*256)+ch;

					int p = 0;

					while(comprTable[c][p] != 0)
					{
						buffer[pos++] = comprTable[c][p++];
					}
				}
				else
				{
					AddCode(buffer, ch, pos);
					pos+=4;
				}
			}
			if(activeNode->type == TYPE_CODE)
			{
				// only add a bracketed code if the current code is not a compression code
				if(curcode != 0x15 && curcode != 0x16 && curcode != 0x17)
				{
					AddCode(buffer, ch, pos);
					pos+=4;
				}
			}



			// if the current node is a first-byte code with value 2, terminate text block
			if(activeNode->level == 0 && activeNode->value == 0x02)
				break;


			// if the active node is a terminator node,  stop stuff
			if(activeNode->isTerminator)
			{
				opcode = false;

				// reset the active node so we start from the beginning
				activeNode = &ccTable;
			}
		}
		else
		{
			// if the current byte is not a control code,
			// add it to the string as-is
			buffer[pos++] = ch;
		}
	}

	buffer[pos++] = 0;

	char* newstr = new char[pos];
	for(i = 0; i < pos; i++)
		newstr[i] = buffer[i];

	newstr[pos] = 0;

	return newstr;
}

int SearchNode(CCNode *node, int val)
{
	// note: if there's only one subnode of the specified node,
	// and it's an argument list, return its index regardless of
	// the specified value.

	if(node->num_nodes == 1 && node->nodes[0].type == TYPE_ARGLIST)
	{
		return 0;
	}

	int retVal = -1;
	for(int i = 0; i < node->num_nodes; i++)
	{
		if(node->nodes[i].value == val)
			retVal = i;
	}

	return retVal;
}


bool isCC(char ch)
{
	bool retVal = false;

	if(ch >= 0 && ch <= 0x1F)
		retVal = true;

	return retVal;
}

void AddCode(char* str, int val, int pos)
{
	char b[5];
	if(val < 0x10)
		sprintf(b, "[0%X]", val);
	else
		sprintf(b, "[%X]", val);

	str[pos++] = b[0];
	str[pos++] = b[1];
	str[pos++] = b[2];
	str[pos++] = b[3];
}



/*void CreateCompressionTable(FILE* rom, char* tbl[])
{
	fseek(rom, 0x8BE2D, SEEK_SET);

	for(int i = 0; i < 768; i++)
	{
		int ch;
		int pos = 0;

		char buffer[256];

		do
		{
			ch = fgetc(rom);
			buffer[pos++] = ch;
		}while(ch != 0);

		tbl[i] = new char[pos+1];

		for(int j = 0; j < pos+1; j++)
			tbl[i][j] = buffer[j]==0 ? 0 : buffer[j]-0x30;
	}
}*/

void CreateCompressionTable(FILE* rom, char* tbl[])
{

	char buffer[256];
	int adr;
	char ch;
	int pos;

	for(int i = 0; i < 768; i++)
	{
		fseek(rom, 0x8CFED+(i*4), SEEK_SET);

		pos = 0;

		adr = fgetc(rom);
		adr+= fgetc(rom) << 8;
		adr+= fgetc(rom) << 16;

		fseek(rom, SNESToHex(adr), SEEK_SET);
		for(int j = 0; j < 256; j++)
		{
			ch = fgetc(rom);
			buffer[pos++] = ch;
			if(ch == 0) break;
		}

		tbl[i] = new char[pos+1];

		for(j = 0; j < pos+1; j++)
			tbl[i][j] = buffer[j]==0 ? 0 : buffer[j]-0x30;
	}
	printf("Created compression table!\n\n");
}


void CreateCCTable(CCNode *table)
{
	FILE* file;
	char buffer[1024];
	unsigned char codestr[1024];
	unsigned char nodestr[1024];
	unsigned char descstr[1024];
	int bpos = 0;
	int pos = 0;
	int len = 0;

	file = fopen("codelist.txt", "r");

	if(file)
	{
		fseek(file, 0, SEEK_SET);

		while(!feof(file))
		{
			fgets(buffer, 1024, file);

			while(buffer[bpos] != ',' && bpos < 1024)
			{
				if(buffer[bpos] != ' ')
				{
					codestr[pos++] = buffer[bpos];
				}
				bpos++;
			}
			codestr[pos] = 0;
			len = pos;
			pos = 0;
			bpos++;

			while(buffer[bpos] != ';' && bpos < 1024)
			{
				if(buffer[bpos] != '"')
				{
					descstr[pos++] = buffer[bpos];
				}
				bpos++;
			}
			descstr[pos] = 0;
			bpos = pos = 0;

			//printf("%s\t:\t", codestr);

			// now convert the string into a series of single bytes
			for(int i = 0; i < len; i)
			{
				char t[3];
				t[0] = codestr[i++];
				t[1] = codestr[i++];
				t[2] = '\0';

				nodestr[pos++] = (unsigned char)strtoul(t, 0, 16);
			}
			nodestr[pos] = 0;
			len = pos;
			pos = 0;


			//printf("%s\n", nodestr);

			// now loop through the code string and add a series of nodes

			int prev = 0;
			bool arg_counting = false;
			int arg_count = 0;
			int args[32];

			CCNode* activeNode = table;
			int nodeLevel = 0;

			for(i = 0; i < len; i++)
			{
				//printf("%X: ", nodestr[i]);
				// if we encounter an argument value (0xAA, 0xBB, etc.),
				// start counting arguments

				// stop counting when the last code no is no longer equal to the current
				if(arg_counting)
				{
					if(prev != nodestr[i])
					{
						//printf("STOP! %d arguments!\n", arg_count);

						// now add latent node
						activeNode = AddNode(activeNode, nodeLevel, nodestr[i-1], TYPE_ARGLIST, arg_count, false, false);
						nodeLevel++;
						arg_count = 0;
						arg_counting = false;
					}
				}
				if((nodestr[i] - 0xAA)%0x11 == 0 && nodestr[i] != 0 && !arg_counting)
				{
					//printf("COUNTING ARGUMENTS!\n");
					arg_counting = true;
				}

				if(arg_counting)
				{
					// if we're still counting arguments, increment the argument count
					// and add the current value to the argument list
					args[arg_count] = nodestr[i];
					arg_count++;
				}
				else
				{
					// add default code node here
					activeNode = AddNode(activeNode, nodeLevel, nodestr[i], TYPE_CODE, NULL, false, (i+1 == len));
					nodeLevel++;
					//printf("\n");
				}
				
				
				prev = nodestr[i];
			}
			if(arg_counting)
			{
				// if we're still counting arguments, tidy up
				
				//printf("STOP! %d arguments!\n", arg_count);

				AddNode(activeNode, nodeLevel, nodestr[i-1], TYPE_ARGLIST, arg_count, false, true);

				arg_counting = false;
				arg_count = 0;
			}

			//printf("\n");
		}
	}
	printf("Created control node network!\n\n");
}


CCNode* AddNode(CCNode *node, int level, int value, int type, int num_arg, bool multi, bool terminate)
{
	// Adds a subnode with specified properties to the provided node

	// if a node of the specified value already exists, return a pointer to it
	for(int i = 0; i < node->num_nodes; i++)
	{
		if(node->nodes[i].value == value)
			return &node->nodes[i];
	}


	CCNode *temp = new CCNode[node->num_nodes];

	for(i = 0; i < node->num_nodes; i++)
		temp[i] = node->nodes[i];

	node->nodes = new CCNode[node->num_nodes+1];

	for(i = 0; i < node->num_nodes; i++)
		node->nodes[i] = temp[i];

	int n = node->num_nodes;

	node->nodes[n].level = level;
	node->nodes[n].value = value;
	node->nodes[n].type = type;
	node->nodes[n].num_arg = num_arg;
	node->nodes[n].arg_multiplier = multi;
	node->nodes[n].isTerminator = terminate;

	node->nodes[n].num_nodes = 0;

	delete temp;
	node->num_nodes++;

	// return a pointer to the node we just created
	return &node->nodes[n];

}



unsigned int HexToSNES(unsigned int address)
{
   // Converts good-ol regular hex addresses into the SNES-style addresses.
   // In this case, it's assumed the ROM is a hi-rom with a header.

   unsigned int retVal;

   retVal = (address + 0xc00000) - 0x200;

   return retVal;
}

unsigned int SNESToHex(unsigned int address)
{
   // Converts an SNES-style address (hi-rom, with header) to a nice
   // handy hex address.

   unsigned int retVal;

   retVal = (address - 0xc00000) + 0x200;

   return retVal;
}